 
/**
 * Painter - Base
 * @param {String}className The name of the painter class - GXRoundRectPainter default
 */
function Painter(className = "GXRoundRectPainter")
{
	var painter = new BasePainter();   
	painter.nsClass = className;
	return painter;
} 

/**
 * TemplatePainter
 * @param {String}template the name of the markup dictionary
 */
function TemplatePainter(template)
{
	var painter = Painter()
	delete painter.nsClass; 
	return painter.setValues({
		"+" : template,
	});
}

/**
 * LabelPainter
 * @param {String}title default string value to paint
 */    
function LabelPainter(title)
{ 
	var painter = Painter("GXStringPainter").color(Colors.controlFill); 
	painter.gxDefaultValue = title; 
	return painter;
}

/**
 * ImagePainter
 * @param {String}imageName name of the image to paint
 */    
function ImagePainter(imageName)
{ 
	return Painter("GXImagePainter").setValues({
		gxImage : imageName,
		gxTintColor: null
	});
}
 
/**
 * Shadow
 * @param {object}color Shadow color object
 * @param {number}x x offset
 * @param {number}y y offset
 * @param {number}radius shadow radius
 */   
function Shadow(color,x,y,radius)
{
	return {
		nsClass: "NSShadow",
		"gxShadowColor" : color,
		"gxBlurRadius" : radius,
		"gxShadowOffset" : `{${x},${y}}`,
	}
}  
 
/**
 * TemplateView - TemplateView Wrapper 
 * @param {String}markupTitle, The name of the object as it is in the markup dictionary
 */  
function TemplateView(markupTitle)
{
	var view = View()
	delete view.nsClass;
	delete view.gxChildViews;
	view["+"] = markupTitle; 
	return view;
} 

/**
 * VList - Vertical GXListView Wrapper 
 * @param {Array} views - Views to host in list
 */ 

function VList(views = [])
{  
	var listView = new BaseListView();
	listView.gxVertical = true; 
	listView.gxCollapsible = false;
	for (var i in views)
    { 
		var view = views[i];    
		listView.addSubView(view.makeView());     
    }  
	return listView;
} 


/**
 * HList - Vertical GXListView Wrapper 
 * @param {Array} views - Views to host in list
 */ 

function HList(views = [])
{  
	var listView = new BaseListView();
	listView.gxVertical = false;
	for (var i in views)
    { 
		var view = views[i];    
		listView.addSubView(view);     
    }  
	return listView;
} 

/**
 * Label 
 * @param {String} defaultValue - Label text
 */

function Label(defaultValue)
{  
	return new BaseLabelView(defaultValue).ignoreAX();
} 

function Color(color = Colors.white)
{ 
	return View().backPainter(Painter().fill(color)).fillWidth().fillHeight().keyPath("");
}

function AnchorView()
{ 
    return new BaseView().setValues({
		isAnchorView: true,
		nsClass: "GXLayoutGuide",
		gxLayerZPosition: -1000, 
		gxViewNameSeed: "_ANCHOR_",
		gxIntrinsicHeight: -1,
		gxIntrinsicWidth: -1,
    });
}
 
function View(nsClass = "GXDrawView")
{   
	var view = new BaseView();
    return view.className(nsClass);
}

function Layer(nsClass = "GXDrawLayer")
{   
	return new BaseLayer().className(nsClass)
}

function LabelLayer(string = "")
{ 
	var layer = Layer("GXStringLayer")
		.setValue(string,"gxDefaultValue")
		.forePainter(LabelPainter(string).color(Colors.tileTint));
	layer.fontColor = function(color)
	{ 
		return this.painterByMutating("gxForePainter",(painter) => {
			return painter.color(color);
		})
	};
	layer.fontSize = function(size)
	{ 
		return this.painterByMutating("gxForePainter",(painter) => {
			return painter.fontSize(size);
		})
	};
	layer.textAlign = function(x,y)
	{ 
		return this.painterByMutating("gxForePainter",(painter) => {
			return painter.align(x,y);
		})
	};
	return layer;
}

function LayerView(layers = [])
{
	return View("GXLayerHostingView").setValues({
		gxLayout: [
			"H:|[rootLayer]|",
			"V:|[rootLayer]|",	 
		],
		gxAutoLoadConstraints: true,
		gxRootLayer: Layer().viewName("rootLayer").subLayers(layers).setValue(false, "gxNeedsDisplayOnResize")
	})
}

function AnimationLayer(asset, tint, duration = 1, loopStart = -1)
{
	return Layer("GXCaarLayer") 
		.fillFrame() 
		// .setValue(true, "gxNeedsDisplayOnResize")
		.caarPainter(CaarPainter(asset, tint, duration, loopStart)); 
}

function CaarPainter(asset, tint, duration = 1, loopStart = -1)
{
	return Painter("GXCaarPainter").setValues({
		gxAsset : asset,
		gxTintColor : tint, 
		gxLoopStart : loopStart,
		gxAssetDuration: duration,  
	});
}
 
function AnimationView(asset, tint, duration = 1, loopStart = -1)
{
	var view = LayerView([
		AnimationLayer(asset, tint, duration, loopStart).fillFrame()
	]);
	return view; 
}

function Padding(size = 10, priority = 1000)
{
    var pad = View("GXLayoutGuide")
	pad.gxLayerZPosition = -1000;
    pad.willAddToSuperview = function(sv)
    {
        sv.stackLayout.push(`${sv.alignmentDirection}:[${this.getViewName()}(${size}@${priority})]`);
    }
    return pad;
}

function Spacer(priority=350)
{
    var spacer = View("GXLayoutGuide"); 
	spacer.gxViewNameSeed = "_SPACER_"
	spacer.gxLayerZPosition = -1000;
    spacer.willAddToSuperview = function(sv)
    { 
        sv.stackLayout.push(`${sv.alignmentDirection}:[${this.getViewName()}(9999@${priority})]`)
    }
    return spacer
}
 

function Divider(color = Colors.divider)
{
	var divider = View().backPainter(Painter().fill(color)).height(1).fillWidth();
	divider.gxViewNameSeed = "_DIVIDER_"
    divider.willAddToSuperview = function(sv)
    { 
		let opp = sv.alignmentDirection == "H" ? "V" : "H"; 
		if (sv['stackLayout'] != null)
		{
			sv.stackLayout.push(`${opp}:[${this.getViewName()}(9999@350)]`)
			sv.stackLayout.push(`${sv.alignmentDirection}:[${this.getViewName()}(1)]`)
		} 
    }
	return divider;
}

function Image(imageName,tintColor)
{
	return View().backPainter(ImagePainter(imageName).tintColor(tintColor))
}

function ImageView(keyPath, tint = Colors.primaryContent)
{
	return View("GXImageView") 
		.keyPath(keyPath)    
		.imagePainter(ImagePainter().tintColor(tint))
}

function SymbolView(symbolName, tint, fontSize = 14.0, fontWeight = 1.0, axDescription = "")
{
	return View("GXSymbolView") 
		.setValues({
			gxSymbolName: symbolName,
			gxFontSize: fontSize,
			gxFontWeight: fontWeight,
			axDescription: axDescription,
			gxTintColor: tint,
			axIsIgnored: axDescription == "",
		})
}

/**
 * Button 
 * @param {String} title - Button title
 */
function Button(title)
{
	var btn = new BaseButtonView(title).flipped(true);
	btn = btn.bordered()
	if (title)
	{
		btn = btn.forePainter(LabelPainter(title))
	}
	return btn
}

/**
 * Button 
 * @param {String} title - Button title
 */
function Checkbox(title)
{
	let btn = new BaseButtonView()  
	return btn
		.height(20)
		.backPainter(ImagePainter("checkbox").tintColor(Colors.tileTint.showPressed()).align(0,0.5))
		.extraPainter(LabelPainter(title).color(Colors.primaryContent.showPressed()).align(0.0,0.5).fontSize(13).offset(20,0))
		.checkBoxRole()
		.flipped()
		.setAXTitle(title)
}


function Slider()
{
	return View("Farrago.FGNoFocusSlider").setValues({ 
		gxRangeMin: 0,
		gxRangeMax: 1,  
		gxTrackPainter: SliderTrack(), 
		gxEndGap: 0,
		gxPreventDragWhileDisabled: true,
		"endGap" : function(gap)
		{
			var cpy = this.copy();
			cpy["gxEndGap"] = gap;
			return cpy;
		}, 
		"hasDetents" : function()
		{
			return this.setValue(TemplatePainter("SliderDetents"), "gxDetentPainter")
		},
		"trackPainter" : function(painter)
		{
			var cpy = this.copy();
			cpy["gxTrackPainter"] = painter;
			return cpy;
		},
		"range" : function(l,h)
		{
			var cpy = this.copy();
			cpy["gxRangeMin"] = l;
			cpy["gxRangeMax"] = h;
			return cpy;
		},
		"readoutPath" : function(keyPath)
		{
			var cpy = this.copy();
			cpy["gxReadoutKeyPath"] = keyPath;
			return cpy; 
		},
		"sliderStyle" : function(style = 0)
		{
			var cpy = this.copy();
			cpy["gxSliderTrackStyling"] = style;
			return cpy; 
		},
		"vertical" : function()
		{
			return this.setValue(true, "gxVertical")
		}, 
		gxDetents: "-1, 0, 1",
		gxTrackMidRatio: 0.0,
		gxSliderTrackStyling: 0,
		gxKeystrokeStepSize: 0.01, 
		gxIgnoreScrollWheel: false,
		gxIgnoreDoubleClick: true,
	}) 
	.endGap(10) 
	.knobPainter(RoundKnob())
	.knobPainterExtra(null)
}


function PopupButton(className = "FGPopupButton", isDropDown = false, extraSize = 20)
{    
	return Button()
		.className(className || "FGPopupButton")
		.intrinsicSize(150,28)
		.backPainter(
			CocoaButtonBack()
		) 
		.forePainter(
			LabelPainter() 
				.tintColor(Colors.primaryContent)
				.fontSize(13)
				.color(Colors.primaryContent)
				.align(0,0.5)
				.monospacedDigits(false)
				.insets(14,0,26,0)
		)
		.iconPainter(ImagePainter().tintColor(Colors.primaryContent.addStyle("disabled", Colors.secondaryContent)).align(0,0.5).insets(8,0,0,0))
		.flipped()
		.extraPainter(Painter()
			.fill(Colors.invisible.addStyle("key",Colors.tileTint.showPressed())) 
			.inset(6)
			.size(extraSize + 8,extraSize) 
			.shadow(MainThemeColor(Colors.invisible, Colors.black.withAlpha(0.4)).brightness(0.1),0,0,1)
			.radius(4))
		.setValues({
			"gxArrowPainter" : ImagePainter(isDropDown ? "dropdown.arrow" : "popup.arrows")
									.offset(0,0) 
									.tintColor(Colors.fontColor.addStyle("key", MainThemeColor(Colors.white.addStyle("8", Colors.white), Colors.white.addStyle("8", Colors.black))).showPressed()), 
			"gxDefaultStyles" : ["0"]
		})
}

function KeyPopupButton(className = "GXPopupMenuView")
{
    return PopupButton()
		.className(className)
		.height(22)
        .forePainter(LabelPainter().align(0.0,0.5).color(Colors.white).fontSize(13).offset(10,0))
		.backPainter(Painter().radius(Constants.buttonRadius).fill(Colors.key.showPressed()).stroke(1, Colors.contrast))
		.drawArrows("caret", Colors.white)
    
} 

function Stack()
{    
    return new BaseStackView();
}   

function VStack(views, space = 0)
{
	var allViews = [];
	for (var v in views)
	{
		allViews.push(views[v]);
		if (space > 0)
		{
			allViews.push(Padding(space))
		} 
	}
	return VStack.private_stack(allViews)
}

VStack.private_stack = function(views) 
{
	var stack = Stack();        
    stack.alignmentDirection = "V";
   
    for (var i in views)
    { 
        var view = views[i];  
         
        if (i > 0)
        { 
            stack.stackLayout.push(`V:${views[i-1].bottomAnchor()}-(0)-${view.topAnchor()}`); 
        }
        else 
        {  
            stack.stackLayout.push(`V:${stack.topAnchor()}-(0)-${view.topAnchor()}`);  
        } 

        stack.addSubView(view);     
        stack.stackLayout.push(`H:${stack.leftAnchor()}-(>=0)-${view.leftAnchor()}`);
        stack.stackLayout.push(`H:${view.rightAnchor()}-(>=0)-${stack.rightAnchor()}`);
    }
    
    if (views.length > 0)
    {
        stack.stackLayout.push(`V:${views.slice(-1)[0].bottomAnchor()}-(0)-${stack.bottomAnchor()}`);
    } 

    return stack.align();
} 

function ZStack(views = [])
{
    var stack = new BaseZStackView();   
    
    for (var i in views)
    { 
        var view = views[i];   
		stack.addSubView(view);  
		
		stack.stackLayout.push(`H:${stack.leftAnchor()}-(>=0)-${view.leftAnchor()}`);
		stack.stackLayout.push(`H:${view.rightAnchor()}-(>=0)-${stack.rightAnchor()}`);  
		stack.stackLayout.push(`V:${stack.topAnchor()}-(>=0)-${view.topAnchor()}`);
        stack.stackLayout.push(`V:${view.bottomAnchor()}-(>=0)-${stack.bottomAnchor()}`);  
	}       
    return stack;
} 

function HStack(views, space = 0)
{
	var allViews = [];
	for (var v in views)
	{
		allViews.push(views[v]);
		if (space > 0 && v != (views.length - 1))
		{
			allViews.push(Padding(space))
		} 
	}
	return HStack.private_stack(allViews)
} 

HStack.private_stack = function(views) 
{
	var stack = Stack();
    stack.alignmentDirection = "H";  
    
    for (var i in views)
    { 
        var view = views[i];  
        if (i > 0)
        { 
            stack.stackLayout.push(`H:${views[i-1].rightAnchor()}-(0)-${view.leftAnchor()}`); 
        }
        else 
        {  
            stack.stackLayout.push(`H:${stack.leftAnchor()}-(0)-${view.leftAnchor()}`);  
		} 
		stack.addSubView(view);   
        stack.stackLayout.push(`V:${stack.topAnchor()}-(>=0)-${view.topAnchor()}`);
        stack.stackLayout.push(`V:${view.bottomAnchor()}-(>=0)-${stack.bottomAnchor()}`); 
    } 
    
    if (views.length > 0)
    {
        stack.stackLayout.push(`H:${views.slice(-1)[0].rightAnchor()}-(0)-${stack.rightAnchor()}`);
    }

    return stack.align();
} 

function ScrollView(documentView = View())
{  
	return View().setValues({ 
		"vScroller" : function(scroller) {
			var cpy = this.copy()
			cpy.gxVScrollbar = scroller
			return cpy
		},
		"vScrollWidth" : function(width) {
			var cpy = this.copy()
			cpy.gxVScrollbarWidth = width
			return cpy
		},
		"hScrollWidth" : function(width) {
			var cpy = this.copy()
			cpy.gxHScrollbarWidth = width
			return cpy
		},
		"hScroller" : function(scroller) {
			var cpy = this.copy()
			cpy.gxHScrollbar = scroller
			return cpy
		},
		"autoHidesScrollers" : function(val) {
			var cpy = this.copy()
			cpy.gxAutohidesScrollers = val
			return cpy
		},  
		axIsIgnored: true,
		gxVScrollbarWidth: 12,
		gxHScrollbarHeight: 12,
		gxIntrinsicHeight: -1,
		gxIntrinsicWidth: -1,
		gxAutohidesScrollers: true, 
		gxHidesAnnoyingCornerView: true,
		gxIgnoreScrollingWhenScrollersAreHidden: true,
		gxWantsLayer: true, // Do not delete! Vital for getting layers to look right on older systems. 
		gxDocumentView: documentView
			.fillFrame(399)
			.addLayout("H:|-0-[self]-(>=0@200)-|")
			.addLayout("V:|-0-[self]-(>=200@200)-|") 
	})
	.className("GXScrollView")
	.backColor(Colors.mainBackground)
	.vScroller(VScrollbarView())
	.hScroller(HScrollbarView());
}

function VScrollbarView(fill = Colors.secondaryBackground, frame = Colors.mainBackground)
{
	return View("GXScroller").setValues({ 
		gxKnobPainter: Painter()
			.insets(4,0,0,0)
			.radius(3)
			.fill(fill.addStyle("hover", MainThemeColor(fill.lightMode.brightness(0.95), fill.darkMode.brightness(1.3))))
			.stroke(0,fill.brightness(1.3))
			.offset(-2,0)
			.inset(2),
		gxMinKnobLength: 1, // avoid resizing that happens in GXScroller
		axIsIgnored: true,
	}) 
}

function HScrollbarView(fill = Colors.secondaryBackground, frame = Colors.mainBackground)
{
	return View("GXScroller").setValues({ 
		gxKnobPainter: Painter()
			.insets(0,4,0,0)
			.radius(3)
			.fill(fill.addStyle("hover", MainThemeColor(fill.lightMode.brightness(0.95), fill.darkMode.brightness(1.3))))
			.stroke(0,fill.brightness(1.3))
			.inset(2),
		gxMinKnobLength: 1, // avoid resizing that happens in GXScroller
		axIsIgnored: true,
	}) 
}
 
function Window(title, contentView)
{ 
	var base = new Base()
	base.nsClass = "GXWindow"
	base.gxWindowTitle = title
	base.gxSavedWindowStateKey = title + "state" 
	base.gxFullSizeContentView = true 
	base.gxWindowMaxSize = "8000,8000";
	base.gxContentView = (contentView.nsClass == "GXDrawView" || contentView.nsClass == null) ? contentView.className("GXSettingsView") : contentView;
	return base
} 

function Popover(contentView = View())
{
	var base = new Base()
	base.nsClass = "GXPopover";
	base.gxInvisibleContentView = true,
	base.gxLayout = [
		"H:|-10-[self]-10-|",
		"V:|-10-[self]-10-|",  
		"H:|-10-[content]-10-|",
		"V:|-10-[content]-10-|",  
	]; 
	base.gxStrokeColor = MainThemeColor(Colors.white.brightness(0.9), Colors.white.brightness(0.30)).addStyle("increasedContrast", Colors.contrastBorder);
	base.gxChildViews = {
		"content" : contentView.viewName("content")
	};
	return base;
}   

function MenuItem(title = "", value = title)
{
	return new View().setValues({
		"gxTitle" : title,
		"gxValue" : value
	})
}

function merge(obj1, obj2) {
	var merged = { };
	Object.assign(merged, obj1, obj2);
    return merged;
};

// // A function for merging dictionaries
// Object.prototype.mergeDict = function(other) {
// 	return Object.assign({}, this, other);
// }